#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "fb2parser.h"


static void parseAuthor      (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseAnnotation  (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseP           (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseStanza      (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseSection     (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseTitleInfo   (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseTitle       (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseDescription (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parsePoem        (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseCite        (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseStyle       (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseEpigraph    (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);
static void parseBody        (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb);

static void bufferAppend(const char *data, int len, FB2Content *fb);
static void addBookmark (int start, int end, FB2Content *fb);


/*     */

static void
bufferAppend(const char *data, int len, FB2Content *fb)
{
  char **buffer;
  size_t *buffer_size;
  int *current_index;

  if (!data) return;

  /*     */
  if (fb->current_buffer == DESCRIPTION_BUFFER) {
    buffer = &(fb->description);
    buffer_size = &(fb->description_buffer_size);
    current_index = &(fb->description_current_index);
  } else {
    buffer = &(fb->text);
    buffer_size = &(fb->text_buffer_size);
    current_index = &(fb->text_current_index);
  }

  /* realloc buffer */
  if (*current_index+len >= *buffer_size) {
    while (1) {
      *buffer_size = *buffer_size*2;
      *buffer = (char *) realloc(*buffer, *buffer_size);
      if (*current_index+len < *buffer_size) {
	break;
      }
    }
  }

  strncpy((*buffer) + *current_index, data, len);
  *current_index = *current_index+len;

  return;
}

/*   */
static void
addBookmark (int start, int end, FB2Content *fb)
{
  FB2Bookmark *bm;
  int i;

  if (fb->current_buffer == TEXT_BUFFER) {
    for (i=0; i<BUF_SIZE-1; i++) {
      if (fb->bookmarks[i] == NULL) {
	bm = (FB2Bookmark *) malloc(sizeof(FB2Bookmark));
	bm->start = start;
	bm->end = end;
	fb->bookmarks[i] = bm;
	fb->bookmarks[i+1] = NULL;
	fb->num_bookmarks++;
	break;
      }
    }
  }
  return;
}


static void
parseAuthor (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{

  xmlChar *content;
  cur = cur->children;
  while (cur != NULL) {

    if ( !xmlStrcmp(cur->name, (const xmlChar *)"first-name")
	 || !xmlStrcmp(cur->name, (const xmlChar *)"middle-name")
	 || !xmlStrcmp(cur->name, (const xmlChar *)"last-name")
	 || !xmlStrcmp(cur->name, (const xmlChar *)"nickname")) {

      content = xmlNodeGetContent(cur->children);
      if (content) {
	int len = strlen(fb->author);
	if (len > 0) {
	  strncat(fb->author, " ", BUF_SIZE-len);
	  len++;
	}
	strncat(fb->author, content, BUF_SIZE-len);
      }

      xmlFree(content);
    }

    cur = cur->next;
  }
  return;
}

static void
parseAnnotation (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  cur = cur->children;

  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"poem")) {
      parsePoem(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cite")) {
      parseCite(doc, cur, fb);

    }

    cur = cur->next;
  }
  return;
}


static void
parseTitleInfo (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{

  xmlChar *content;
  cur = cur->children;
  while (cur != NULL) {
    //printf("title-info: %s\n", cur->name);

    if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"genre")) {
      content = xmlNodeGetContent(cur->children);

      if(content) {
	/* add genre to categories */
	int i;
	for (i=0; i<BUF_SIZE-1; i++) {
	  if (fb->categories[i] == NULL) {
	    fb->categories[i] = content;
	    fb->categories[i+1] = NULL;
	    fb->num_categories++;
	    break;
	  }
	}
      }

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"book-title")) {
      content = xmlNodeGetContent(cur->children);
      if (content)
	strncpy(fb->name, content, BUF_SIZE);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"author")) {
      parseAuthor(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"annotation")) {
      parseAnnotation(doc, cur, fb);

    }
    cur = cur->next;
  }
  return;
}


static void
parseTitle (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  int start;

  bufferAppend("\n", 1, fb);

  start = fb->text_current_index;

  cur = cur->children;

  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    }
    cur = cur->next;
  }

  addBookmark(start, fb->text_current_index-1, fb);

  bufferAppend("\n", 1, fb);

  return;
}


static void
parseDescription (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  cur = cur->children;
  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"title-info")) {
      parseTitleInfo(doc, cur, fb);

    }
    cur = cur->next;
  }
  return;
}


static void
parseEpigraph (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  xmlChar *content;

  cur = cur->children;
  while (cur != NULL) {

    if (xmlNodeIsText(cur)) {
      content = xmlNodeGetContent(cur);
      bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cite")) {
      parseCite(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"poem")) {
      parsePoem(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    }

    cur = cur->next;
  }
  return;
}


static void
parseP (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  xmlChar *content;
  cur = cur->children;

  bufferAppend("\t", 1, fb);

  while (cur != NULL) {

    if (xmlNodeIsText(cur)) {
      content = xmlNodeGetContent(cur);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"style")) {
      parseStyle(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"strong")) {
      content = xmlNodeGetContent(cur->children);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"emphasis")) {
      content = xmlNodeGetContent(cur->children);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    }

    cur = cur->next;
  }

  bufferAppend("\n", 1, fb);
  return;
}

static void
parsePoem (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  cur = cur->children;
  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"title")) {
      parseTitle(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"epigraph")) {
      parseEpigraph(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"stanza")) {
      bufferAppend("\n", 1, fb);
      parseStanza(doc, cur, fb);
    }

    cur = cur->next;
  }
  bufferAppend("\n", 1, fb);
  return;
}

static void
parseStanza (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  xmlChar *content;
  cur = cur->children;

  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"v")) {
      content = xmlNodeGetContent(cur->children);
      if (content) {
	bufferAppend("\t", 1, fb);
	bufferAppend(content, strlen(content), fb);
	bufferAppend("\n", 1, fb);
      }
      xmlFree(content);
    }

    cur = cur->next;
  }
  return;
}

static void
parseCite (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  xmlChar *content;
  cur = cur->children;

  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"poem")) {
      parsePoem(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"p") ) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"text-author")) {
      content = xmlNodeGetContent(cur->children);
      if (content) {
	bufferAppend("\t", 1, fb);
	bufferAppend(content, strlen(content), fb);
	bufferAppend("\n", 1, fb);
      }
      xmlFree(content);

    }

    cur = cur->next;
  }
  return;
}


static void
parseStyle (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{
  xmlChar *content;
  cur = cur->children;

  while (cur != NULL) {

    if (xmlNodeIsText(cur)) {
      content = xmlNodeGetContent(cur);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"style")) {
      parseStyle(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"strong")) {
      content = xmlNodeGetContent(cur->children);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"emphasis")) {
      content = xmlNodeGetContent(cur->children);
      if (content)
	bufferAppend(content, strlen(content), fb);
      xmlFree(content);

    }

    cur = cur->next;
  }
  return;
}


static void
parseSection (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{

  //xmlChar *content;

  cur = cur->children;
  while (cur != NULL) {

    if (!xmlStrcmp(cur->name, (const xmlChar *)"title")) {
      parseTitle(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"section")) {
      parseSection(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"poem")) {
      parsePoem(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cite")) {
      parseCite(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"epigraph")) {
      parseEpigraph(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"p")) {
      parseP(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"subtitle")) {
      int start = fb->text_current_index;
      parseP(doc, cur, fb);
      addBookmark(start, fb->text_current_index, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"empty-line")) {
      bufferAppend("\n", 1, fb);

    }

    cur = cur->next;
  }
  return;
}


static void
parseBody (xmlDocPtr doc, xmlNodePtr cur, FB2Content *fb)
{

  //xmlChar *content;
  cur = cur->children;
  while (cur != NULL) {
    //printf("body: %s\n", cur->name);

    if (!xmlStrcmp(cur->name, (const xmlChar *)"section")) {
      parseSection(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"title")) {
      parseTitle(doc, cur, fb);

    }

    cur = cur->next;
  }
  return;
}

FB2Content *
parseFile(char *filename)
{

  xmlDocPtr doc;
  xmlNodePtr cur;
  FB2Content *fb;


  doc = xmlParseFile(filename);

  if (doc == NULL ) {
    fprintf(stderr, "Document not parsed successfully. \n");
    return NULL;
  }

  cur = xmlDocGetRootElement(doc);

  if (cur == NULL) {
    fprintf(stderr, "empty document\n");
    xmlFreeDoc(doc);
    return NULL;
  }

  if (xmlStrcmp(cur->name, (const xmlChar *) "FictionBook")) {
    fprintf(stderr, "document of the wrong type, root node != FictionBook");
    xmlFreeDoc(doc);
    return NULL;
  }


  fb = (FB2Content *) malloc(sizeof(FB2Content));

  /* initialize buffers */
  fb->text = (char *) malloc(1024);
  fb->text_buffer_size = 1024;
  fb->text_current_index = 0;
  fb->description = (char *) malloc(1024);
  fb->description_buffer_size = 1024;
  fb->description_current_index = 0;

  *fb->author = '\0';
  *fb->name = '\0';

  fb->categories[0] = NULL;
  fb->num_categories = 0;
  fb->bookmarks[0] = NULL;
  fb->num_bookmarks = 0;

  cur = cur->children;
  while (cur != NULL) {
    //printf("root: %s\n", cur->name);

    if (!xmlStrcmp(cur->name, (const xmlChar *)"description")) {
      fb->current_buffer = DESCRIPTION_BUFFER;
      parseDescription(doc, cur, fb);

    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"body")) {
      fb->current_buffer = TEXT_BUFFER;
      parseBody(doc, cur, fb);

    }

    cur = cur->next;
  }
  xmlFreeDoc(doc);

  fb->current_buffer = DESCRIPTION_BUFFER;
  bufferAppend("\0", 1, fb);

  fb->current_buffer = TEXT_BUFFER;
  bufferAppend("\0", 1, fb);

  return fb;
}

void
freeFB2 (FB2Content *fb)
{
  int i;

  for (i=0; i<BUF_SIZE-1; i++) {
    if (fb->categories[i] == NULL)
      break;
    xmlFree(fb->categories[i]);
  }

  for (i=0; i<BUF_SIZE-1; i++) {
    FB2Bookmark *bm = fb->bookmarks[i];
    if (bm == NULL)
      break;
    free(bm);
  }

  free(fb->description);
  free(fb->text);

  return;
}

int
main(int argc, char **argv)
{
  char *docname;
  FB2Content *fb;
  int i;

  if (argc <= 1) {
    printf("Usage: %s docname\n", argv[0]);
    return 1;
  }

  docname = argv[1];
  fb = parseFile(docname);

  /******** test ********/

  printf("author: %s\n", fb->author);
  printf("name: %s\n", fb->name);

  printf("categories:\n");

  for (i=0; i<BUF_SIZE-1; i++) {
    if (fb->categories[i] == NULL) {
      break;
    }
    printf("[%d] %s\n", i, fb->categories[i]);
  }

  printf("-- bookmarks --\n");

  for (i=0; i<BUF_SIZE-1; i++) {
    FB2Bookmark *bm;
    int j;
    bm = fb->bookmarks[i];

    if (bm == NULL) {
      break;
    }

    for (j = bm->start; j < bm->end; j++)
      printf("%c", fb->text[j]);

    printf("\n");
  }

  printf("-- description\n");

  printf("%s\n", fb->description);

  printf("-- text --\n");

  printf("%s\n", fb->text);

  freeFB2(fb);

  return 0;
}


